/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"
#include <algorithm>
#include <vector>
#include <memory>

class SortExecutor : public AbstractExecutor { // by march7 20240528
private:
    std::unique_ptr<AbstractExecutor> prev_; // 前一个执行器
    ColMeta cols_; // 需要排序的列
    bool is_desc_; // 排序方式
    std::vector<std::unique_ptr<RmRecord>> tuples_; // 存储排序后的元组
    size_t tuple_num; // 当前元组索引
    bool already_sorted_; // 输入数据是否已经有序 // for sort merge join, march7 20240529

public:
    SortExecutor(std::unique_ptr<AbstractExecutor> prev, TabCol sel_cols, bool is_desc, bool is_sorted)
        : prev_(std::move(prev)), is_desc_(is_desc), already_sorted_(is_sorted), tuple_num(0) {
        cols_ = *get_col(prev_->cols(), sel_cols);
    }

    void beginTuple() override {
        // 获取所有元组并排序
        // std::cout << "tab name " << cols_.tab_name << " sort col name " << cols_.name << std::endl;
        tuples_.clear();
        prev_->beginTuple();
        while (!prev_->is_end()) {
            auto tuple = prev_->Next();
            tuples_.push_back(std::move(tuple));
            prev_->nextTuple();
        }
        if (!already_sorted_) {
            printf("not already sorted\n");
            std::sort(tuples_.begin(), tuples_.end(), [this](const std::unique_ptr<RmRecord>& a, const std::unique_ptr<RmRecord>& b) {
                auto a_val = a->data + cols_.offset;
                auto b_val = b->data + cols_.offset;
                int cmp = ix_compare(a_val, b_val, cols_.type, cols_.len);
                return is_desc_ ? cmp > 0 : cmp < 0;
            });
        } else {
            printf("already after sortmerge join, no need sort again\n");
        }
        tuple_num = 0;
    }

    void nextTuple() override {
        if (tuple_num < tuples_.size()) {
            tuple_num++;
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if (tuple_num < tuples_.size()) {
            return std::make_unique<RmRecord>(*tuples_[tuple_num]);
        }
        return nullptr;
    }

    Rid &rid() override {
        return _abstract_rid;
    }

    bool is_end() const override {
        return tuple_num >= tuples_.size();
    }

    size_t tupleLen() const override {
        return prev_->tupleLen();
    }

    const std::vector<ColMeta>& cols() const override {
        return prev_->cols();
    }

    std::string getType() override {
        return "SortExecutor";
    }

};
