/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
   private:
    std::unique_ptr<AbstractExecutor> left_;    // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;   // 右儿子节点（需要join的表）
    size_t len_;                                // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                 // join后获得的记录的字段

    std::vector<Condition> fed_conds_;          // join条件
    bool isend;

    Rid _abstract_rid;
    std::unique_ptr<RmRecord> left_record_;     // 当前左表记录
    std::unique_ptr<RmRecord> right_record_;    // 当前右表记录

   public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, 
                            std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto &col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        isend = false;
        fed_conds_ = std::move(conds);

    }

    void beginTuple() override { //zmjjkk 2024/5/22
        // 初始化左右表的扫描器
        left_->beginTuple();
        right_->beginTuple();

        // 获取左表的第一个记录
        left_record_ = left_->Next();
        
        // 循环直到找到第一个满足条件的左右记录对
        while (!left_->is_end()) {
            right_record_ = right_->Next();

            while (!right_->is_end()) {
                if (checkConditions()) {  // 检查当前记录对是否满足条件
                    _abstract_rid = right_->rid();
                    return;
                }
                right_->nextTuple();
                right_record_ = right_->Next();  // 获取右表的下一个记录
            }
            left_->nextTuple();
            left_record_ = left_->Next();  // 获取左表的下一个记录
            right_->beginTuple();  // 重新开始右表的扫描

        }

        // 如果没有找到任何满足条件的记录对，标记为结束
        isend = true;
    }

    void nextTuple() override { //zmjjkk 2024/5/22
        // if (!left_record_) {
        //     isend = true;
        //     return;
        // }
        //same begintuple()
        right_->nextTuple();
        while (!left_->is_end()) {
            right_record_ = right_->Next();
            while (!right_->is_end()) {
                if (checkConditions()) {
                    _abstract_rid = right_->rid();
                    return;
                }
                right_->nextTuple();
                right_record_ = right_->Next();
            }
            left_->nextTuple();
            left_record_ = left_->Next();  // 获取左表的下一个记录
            right_->beginTuple();  // 重新开始右表的扫描
        }
        isend = true;
    }

    std::unique_ptr<RmRecord> Next() override { //zmjjkk 2024/5/22
        if (isend) {
            return nullptr;
        }
        std::unique_ptr<RmRecord> joined_record = std::make_unique<RmRecord>(len_);
        std::memcpy(joined_record->data, left_record_->data, left_->tupleLen());
        std::memcpy(joined_record->data + left_->tupleLen(), right_record_->data, right_->tupleLen());
        return joined_record;
    }

    Rid &rid() override {
        return _abstract_rid;
    }

    bool checkConditions() { //zmjjkk 2024/5/22
        bool check = true;
        for (const auto &cond : fed_conds_) {
            const auto &lhs_col_t = cond.lhs_col;
            const auto &rhs_col_t = cond.rhs_col;
            auto lhs_col = get_col(left_->cols(), lhs_col_t);
            auto rhs_col = get_col(right_->cols(), rhs_col_t);
            char* lhs_data = left_record_->data +lhs_col->offset;
            char* rhs_data = right_record_->data +rhs_col->offset;
            ColType l_type = lhs_col->type;
            ColType r_type = rhs_col->type;
            int cmp = 0;
            if (l_type == r_type) {
                cmp = ix_compare(lhs_data, rhs_data, l_type, lhs_col->len);
            }
            else {
            if (r_type == ColType::TYPE_STRING || l_type == ColType::TYPE_STRING) {
                throw IncompatibleTypeError(coltype2str(l_type), coltype2str(r_type));
            }
                double lf = tofloat(lhs_data, lhs_col->type);
                double rf = tofloat(rhs_data, rhs_col->type);
                double diff = lf - rf;
                const double epsilon = 1e-7f;
                if (std::abs(diff) < epsilon) {
                    cmp = 0;
                } else {
                    cmp = (diff < 0) ? -1 : 1;
                }
            }
            switch (cond.op) {
            case OP_EQ:
                check = check && (cmp == 0);
                break;
            case OP_NE:
                check = check && (cmp != 0);
                break;
            case OP_LT:
                check = check && (cmp < 0);
                break;
            case OP_GT:
                check = check && (cmp > 0);
                break;
            case OP_LE:
                check = check && (cmp <= 0);
                break;
            case OP_GE:
                check = check && (cmp >= 0);
                break;
            default:
                throw UnknownOperatorError();
            }
        }
        return check;
    }

   double tofloat(char* data, ColType type) { //zmjjkk 2024/5/22
        switch (type) {
        case TYPE_INT: {
            // 将整型数据转换为 float
            int int_val;
            std::memcpy(&int_val, data, sizeof(int));
            return static_cast<double>(int_val);
        }
        case TYPE_FLOAT: {
            // 将浮点型数据直接转换为 float
            double float_val;
            std::memcpy(&float_val, data, sizeof(double));
            return float_val;
        }
        default:
            throw std::invalid_argument("Unsupported column type");
        }
    }
    
    const std::vector<ColMeta> &cols() const override { return cols_; }

    size_t tupleLen() const override { return len_; }

    bool is_end() const override { return isend; }

};