/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */
// by zmjjkk 20240521
#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
   private:
    std::string tab_name_;              // 表的名称
    std::vector<Condition> conds_;      // scan的条件
    RmFileHandle *fh_;                  // 表的数据文件句柄
    std::vector<ColMeta> cols_;         // scan后生成的记录的字段
    size_t len_;                        // scan后生成的每条记录的长度
    std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同

    Rid rid_;
    std::unique_ptr<RecScan> scan_;     // table_iterator

    SmManager *sm_manager_;

   public:
    SeqScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = std::move(tab_name);
        conds_ = std::move(conds);
        TabMeta &tab = sm_manager_->db_.get_table(tab_name_);
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab.cols;
        len_ = cols_.back().offset + cols_.back().len;
        context_ = context;
        fed_conds_ = conds_;
    }

    void beginTuple() override {
        // 找到第一个符合条件的记录
        scan_ = std::make_unique<RmScan>(fh_);  // 使用具体实现
        while (!scan_->is_end()) {
            rid_=scan_->rid();
            if(!check_conditions(fh_->get_record(rid_, context_))) {
                scan_->next();
            }
            else {
                break;
            }
        }
    }

    void nextTuple() override {
        scan_->next();
        while (!scan_->is_end()) {
            rid_ = scan_->rid();
            std::unique_ptr<RmRecord> record = fh_->get_record(rid_, context_);
            if (check_conditions(record)) {
                return;
            }
            scan_->next();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
         if (scan_->is_end()) {
            return nullptr;
        }
        std::unique_ptr<RmRecord> record = fh_->get_record(rid_, context_);
        return record;
    }

    
    bool check_conditions(const std::unique_ptr<RmRecord> &record) {
        bool check = true;
        for (const auto &cond : conds_) {
            auto lhs_col = get_col(cols_, cond.lhs_col);
            char *lhs_data = record->data + lhs_col->offset;
            char *rhs_data = nullptr;
            ColType rhs_type;
            if(cond.is_rhs_val) {
                rhs_data = cond.rhs_val.raw->data;
                rhs_type = cond.rhs_val.type;
            }else {
                auto rhs_col = get_col(cols_, cond.rhs_col);
                rhs_type = rhs_col->type;
                rhs_data = record->data + rhs_col->offset;
            }
            int cmp = 0;
            if (rhs_type == lhs_col->type) {
                cmp = ix_compare(lhs_data, rhs_data, rhs_type, lhs_col->len);
            }
            else {
                if (rhs_type == ColType::TYPE_STRING || lhs_col->type == ColType::TYPE_STRING) {
                    throw IncompatibleTypeError(coltype2str(lhs_col->type), coltype2str(rhs_type));
                }
                double lf = tofloat(lhs_data, lhs_col->type);
                double rf = tofloat(rhs_data, rhs_type);
                double diff = lf - rf;
                // std::cout<<lf<<std::endl<<rf<<std::endl<<diff<<std::endl;
                const float epsilon = 1e-7f;
                if (std::abs(diff) < epsilon) {
                    cmp = 0;
                } else {
                    cmp = (diff < 0) ? -1 : 1;
                }
            }
            switch (cond.op) {
            case OP_EQ:
                check = check && (cmp == 0);
                break;
            case OP_NE:
                check = check && (cmp != 0);
                break;
            case OP_LT:
                check = check && (cmp < 0);
                break;
            case OP_GT:
                check = check && (cmp > 0);
                break;
            case OP_LE:
                check = check && (cmp <= 0);
                break;
            case OP_GE:
                check = check && (cmp >= 0);
                break;
            default:
                throw UnknownOperatorError();
            }
        }
        return check;
    }
    double tofloat(char* data, ColType type) {
        switch (type) {
        case TYPE_INT: {
            // 将整型数据转换为 float
            int int_val;
            std::memcpy(&int_val, data, sizeof(int));
            return static_cast<double>(int_val);
        }
        case TYPE_FLOAT: {
            // 将浮点型数据直接转换为 float
            double float_val;
            std::memcpy(&float_val, data, sizeof(double));
            return float_val;
        }
        default:
            throw std::invalid_argument("Unsupported column type");
        }
    }

    Rid &rid() override { return rid_; }
    
    const std::vector<ColMeta> &cols() const override { return cols_; }

    size_t tupleLen() const override { return len_; }

    bool is_end() const override { return scan_->is_end(); }

    std::string getType() override {
        return "SeqScanExecutor";
    }

};
