/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

// create by march7 20240529

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "executor_seq_scan.h"
#include "executor_index_scan.h"
#include "index/ix.h"
#include "system/sm.h"
#include <algorithm>
#include <memory>
#include <vector>

// 比较函数，使用 ix_compare 进行字段比较
inline int compare(const char *a, const char *b, ColType type, int col_len) {
    return ix_compare(a, b, type, col_len);
}

// 输出排序结果到文件
void outputSortedResults(const std::vector<std::unique_ptr<RmRecord>>& records, const std::vector<ColMeta>& cols, std::fstream& outfile) {
    // outfile << table_name << ":\n|";
    outfile << "|";
    for (auto& col : cols) {
        outfile << " " << col.name << " |";
    }
    outfile << "\n";
    for (auto& record : records) {
        outfile << "|";
        for (auto& col : cols) {
            std::string col_str;
            char* rec_buf = record->data + col.offset;
            if (col.type == TYPE_INT) {
                col_str = std::to_string(*(int*)rec_buf);
            } else if (col.type == TYPE_FLOAT) {
                char str[20];
                sprintf(str, "%.6lf", *(double*)rec_buf);
                col_str = str;
            } else if (col.type == TYPE_STRING) {
                col_str = std::string((char*)rec_buf, col.len);
                col_str.resize(strlen(col_str.c_str()));
            }
            outfile << " " << col_str << " |";
        }
        outfile << "\n";
    }
}

class SortMergeJoinExecutor : public AbstractExecutor {
private:
    std::unique_ptr<AbstractExecutor> left_;    // 左子节点
    std::unique_ptr<AbstractExecutor> right_;   // 右子节点
    size_t len_;                                // 连接后记录的长度
    std::vector<ColMeta> cols_;                 // 连接后记录的字段
    std::vector<ColMeta>::const_iterator left_col_;  // 连接条件的左字段
    std::vector<ColMeta>::const_iterator right_col_; // 连接条件的右字段
    int op_;                                    // 操作符
    std::vector<std::unique_ptr<RmRecord>> left_sorted_;  // 排序后的左表记录
    std::vector<std::unique_ptr<RmRecord>> right_sorted_; // 排序后的右表记录
    size_t left_index_;                         // 当前左表记录的索引
    size_t right_index_;                        // 当前右表记录的索引
    bool is_end_;                               // 标记连接是否结束
    bool is_output_;                            // if output to file

    // 比较两个记录，返回比较结果
    inline int compareRecords(const std::unique_ptr<RmRecord>& left_record, const std::unique_ptr<RmRecord>& right_record) {
        if (op_ == OP_EQ) {
            char* left_data = left_record->data + left_col_->offset;
            char* right_data = right_record->data + right_col_->offset;
            return compare(left_data, right_data, left_col_->type, left_col_->len);
        } else {
            throw UnknownOperatorError();
        }
        
        return 0;
    }

public:
    SortMergeJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right, std::vector<Condition> conds, bool is_output)
        : left_(std::move(left)), right_(std::move(right)), left_index_(0), right_index_(0), is_end_(false) {
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto& col : right_cols) {
            col.offset += left_->tupleLen();
        }
        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        const auto& cond = conds[0]; // 目前只匹配一个条件，而且必须是等值连接
        left_col_ = get_col(left_->cols(), cond.lhs_col);
        right_col_ = get_col(right_->cols(), cond.rhs_col);
        op_ = cond.op;
        is_output_ = is_output;
    }

    void beginTuple() override {
        left_sorted_.clear();
        right_sorted_.clear();

        std::cout<<"ltype "<<left_->getType()<<" rtyoe "<<right_->getType()<<std::endl;
        if (left_->getType() == "IndexScanExecutor") {
            // 使用索引扫描器读取左表数据
            left_->beginTuple();
            while (!left_->is_end()) {
                left_sorted_.push_back(left_->Next());
                left_->nextTuple();
            }
            // printf("zuobian sao suoyin\n");
        } else {
            // 使用顺序扫描器读取左表数据
            left_->beginTuple();
            while (!left_->is_end()) {
                left_sorted_.push_back(left_->Next());
                left_->nextTuple();
            }
            // 对左表的记录进行排序
            int left_offset = left_col_->offset;
            std::sort(left_sorted_.begin(), left_sorted_.end(), [this, left_offset](const auto& lhs, const auto& rhs) {
                return compare(lhs->data + left_offset, rhs->data + left_offset, left_col_->type, left_col_->len) < 0;
            });
            sleep(1);
            printf("zuobian sao biao\n");
        }

        if (right_->getType() == "IndexScanExecutor") {
            // 使用索引扫描器读取右表数据
            right_->beginTuple();
            while (!right_->is_end()) {
                right_sorted_.push_back(right_->Next());
                right_->nextTuple();
            }
            // printf("youbian sao suoyin\n");
        } else {
            // 使用顺序扫描器读取右表数据
            right_->beginTuple();
            while (!right_->is_end()) {
                right_sorted_.push_back(right_->Next());
                right_->nextTuple();
            }
            // 对右表的记录进行排序
            int right_offset = right_col_->offset;
            std::sort(right_sorted_.begin(), right_sorted_.end(), [this, right_offset](const auto& lhs, const auto& rhs) {
                return compare(lhs->data + right_offset, rhs->data + right_offset, right_col_->type, right_col_->len) < 0;
            });
            sleep(1);
            printf("youbian sao biao\n");
        }

        if (is_output_) {
            std::fstream outfile;
            outfile.open("sorted_results.txt", std::ios::out | std::ios::ate); // clear and rewrite

            // 输出排序后的左表结果
            outputSortedResults(right_sorted_, right_->cols(), outfile);

            // 输出排序后的右表结果
            outputSortedResults(left_sorted_, left_->cols(), outfile);

            outfile.close();
        }

        left_index_ = 0;
        right_index_ = 0;
        is_end_ = left_sorted_.empty() || right_sorted_.empty();
        nextTuple();
    }

    void nextTuple() override {
        // printf("next tuple\n");

        if (!is_end_) {  // 如果连接还未结束

            size_t left_size = left_sorted_.size();
            size_t right_size = right_sorted_.size();

            // printf("end? %d, lindex %d, rindex %d\n", 
            //     is_end_, left_index_, right_index_);

            // 循环直到找到下一个满足条件的记录对或者遍历完左右表
            while (left_index_ < left_size && right_index_ < right_size) {
                int cmp = compareRecords(left_sorted_[left_index_], right_sorted_[right_index_]);
                if (cmp == 0) {
                    // 如果找到满足条件的记录对，返回
                    return;
                } else if (cmp < 0) {
                    // 如果左记录小于右记录，则推进左索引
                    left_index_++;
                } else {
                    // 如果左记录大于右记录，则推进右索引
                    right_index_++;
                }
            }

            // 如果遍历完左表或右表，标记连接结束
            is_end_ = (left_index_ >= left_size || right_index_ >= right_size);

        }
    }

    std::unique_ptr<RmRecord> Next() override {
        // printf("next\n");
        // printf("end? %d, lindex %d, lsize %d, rindex %d, rsize %d\n", 
        //     is_end_, left_index_, left_sorted_.size(), right_index_, right_sorted_.size());
        if (is_end_ || left_index_ >= left_sorted_.size() || right_index_ >= right_sorted_.size()) {
            return nullptr;
        }
        auto joined_record = std::make_unique<RmRecord>(len_);
        std::memcpy(joined_record->data, left_sorted_[left_index_]->data, left_->tupleLen());
        std::memcpy(joined_record->data + left_->tupleLen(), right_sorted_[right_index_]->data, right_->tupleLen());
        // printf("next finished\n");

        // Move to the next record for the next call
        right_index_++;
        nextTuple();

        return joined_record;
    }

    Rid& rid() override {
        static Rid invalid_rid;
        return invalid_rid;
    }

    const std::vector<ColMeta>& cols() const override { return cols_; }

    size_t tupleLen() const override { return len_; }

    bool is_end() const override { return is_end_; }
};
