/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
   private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle *fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager *sm_manager_;

   public:
    UpdateExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context *context) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;
    }

std::unique_ptr<RmRecord> Next() override {
    // 检查是否还有记录需要更新 by jinx 20240523 
    //gpt always giving wrong answer
    if (rids_.empty()) {
        std::cout << "No records to update.\n";
        return nullptr; // 没有需要更新的记录，返回空指针
    }

    // 初始化设置子句中的右值
    for (auto& set_clause : set_clauses_) {
        // 获取设置子句中左值对应列的元数据
        auto lhs_col = tab_.get_col(set_clause.lhs.col_name);
        // 初始化右值的原始数据
        if (lhs_col->type == TYPE_FLOAT && set_clause.rhs.type == TYPE_INT) {
            set_clause.rhs.set_float(static_cast<double>(set_clause.rhs.int_val));
        }
        if (lhs_col->type == TYPE_INT && set_clause.rhs.type == TYPE_FLOAT) {
            set_clause.rhs.set_int(static_cast<int>(set_clause.rhs.float_val));
        }
        set_clause.rhs.init_raw(lhs_col->len);
    }

    // 更新每个记录
    for (auto& rid : rids_) {
        std::cout << "Updating record with RID: "<< rid.page_no << "\n";
        
        // 获取当前记录
        auto rec = fh_->get_record(rid, context_);
        std::cout << "Original record data: " << std::string(rec->data, rec->size) << "\n";

        // 记录一个更新操作到事务中
        RmRecord update_record{ rec->size };
        memcpy(update_record.data, rec->data, rec->size);
          // 更新记录文件中的记录
        for (auto& set_clause : set_clauses_) {
            auto col_meta = tab_.get_col(set_clause.lhs.col_name);
            int offset = col_meta->offset;
            int len = col_meta->len;

            memcpy(update_record.data + offset, set_clause.rhs.raw->data, len);
        }
        // 处理索引删除旧的条目和检查新值是否违反唯一约束
        std::vector<char*> old_keys(tab_.indexes.size());
        std::vector<char*> new_keys(tab_.indexes.size());
        for (size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto& index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
            
            old_keys[i] = new char[index.col_tot_len];
            new_keys[i] = new char[index.col_tot_len];
            
            int offset = 0;
            for (size_t j = 0; j < index.col_num; ++j) {
                memcpy(old_keys[i] + offset, rec->data + index.cols[j].offset, index.cols[j].len);
                memcpy(new_keys[i] + offset, update_record.data + index.cols[j].offset, index.cols[j].len);
                offset += index.cols[j].len;
            }
            
            std::cout << "Old key: " << std::string(old_keys[i], index.col_tot_len) << "\n";
            std::cout << "New key: " << std::string(new_keys[i], index.col_tot_len) << "\n";
            
            // 检查唯一约束
            std::vector<Rid> temp_rids;
            if (memcmp(old_keys[i], new_keys[i], index.col_tot_len) != 0 && ih->get_value(new_keys[i], &temp_rids, context_->txn_) ) {
                std::cout << "Unique constraint violation detected.\n";
                for (size_t k = 0; k <= i; ++k) {
                    delete[] old_keys[k];
                    delete[] new_keys[k];
                }
                throw UniqueConstraintViolationError();
            }
        }

  
        std::cout << "Updated record data: " << std::string(update_record.data, update_record.size) << "\n";

        // 更新索引
        for (size_t i = 0; i < tab_.indexes.size(); ++i) {
            auto& index = tab_.indexes[i];
            auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();

            if (memcmp(old_keys[i], new_keys[i], index.col_tot_len) != 0) {
                ih->delete_entry(old_keys[i], context_->txn_);
                ih->insert_entry(new_keys[i], rid, context_->txn_);
            }

            delete[] old_keys[i];
            delete[] new_keys[i];
        }

        // 更新记录文件中的记录
        fh_->update_record(rid, update_record.data, context_);
        std::cout << "Record updated successfully.\n";
    }

    return nullptr;
}

    

    Rid &rid() override { return _abstract_rid; }
};