/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "sm_manager.h"

#include <sys/stat.h>
#include <unistd.h>

#include <fstream>

#include "index/ix.h"
#include "record/rm.h"
#include "record_printer.h"

/**
 * @description: 判断是否为一个文件夹
 * @return {bool} 返回是否为一个文件夹
 * @param {string&} db_name 数据库文件名称，与文件夹同名
 */
bool SmManager::is_dir(const std::string& db_name) {
    struct stat st;
    return stat(db_name.c_str(), &st) == 0 && S_ISDIR(st.st_mode);
}

/**
 * @description: 创建数据库，所有的数据库相关文件都放在数据库同名文件夹下
 * @param {string&} db_name 数据库名称
 */
void SmManager::create_db(const std::string& db_name) {
    if (is_dir(db_name)) {
        throw DatabaseExistsError(db_name);
    }
    //为数据库创建一个子目录
    std::string cmd = "mkdir " + db_name;
    if (system(cmd.c_str()) < 0) {  // 创建一个名为db_name的目录
        throw UnixError();
    }
    if (chdir(db_name.c_str()) < 0) {  // 进入名为db_name的目录
        throw UnixError();
    }
    //创建系统目录
    DbMeta *new_db = new DbMeta();
    new_db->name_ = db_name;

    // 注意，此处ofstream会在当前目录创建(如果没有此文件先创建)和打开一个名为DB_META_NAME的文件
    std::ofstream ofs(DB_META_NAME);

    // 将new_db中的信息，按照定义好的operator<<操作符，写入到ofs打开的DB_META_NAME文件中
    ofs << *new_db;  // 注意：此处重载了操作符<<

    delete new_db;

    // 创建日志文件
    disk_manager_->create_file(LOG_FILE_NAME);

    // 回到根目录
    if (chdir("..") < 0) {
        throw UnixError();
    }
}

/**
 * @description: 删除数据库，同时需要清空相关文件以及数据库同名文件夹
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
void SmManager::drop_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }
    std::string cmd = "rm -r " + db_name;
    if (system(cmd.c_str()) < 0) {
        throw UnixError();
    }
}

/**
 * @description: 打开数据库，找到数据库对应的文件夹，并加载数据库元数据和相关文件
 * @param {string&} db_name 数据库名称，与文件夹同名
 */
// by march7 20240521
void SmManager::open_db(const std::string& db_name) {
    if (!is_dir(db_name)) {
        throw DatabaseNotFoundError(db_name);
    }

    // 切换到数据库目录
    if (chdir(db_name.c_str()) < 0) {
        throw UnixError();
    }

    // 读取数据库元数据
    std::ifstream ifs(DB_META_NAME);
    ifs >> db_;

    // 打开所有表的记录文件和索引文件
    for (auto& table_entry : db_.tabs_) {
        const std::string& table_name = table_entry.first;
        const TabMeta& tab_meta = table_entry.second;

        // 打开记录文件
        fhs_[table_name] = rm_manager_->open_file(table_name);

        // 打开索引文件
        for (auto& index_entry : tab_meta.indexes) {
            const std::vector<ColMeta>& index_cols = index_entry.cols;
            std::unique_ptr<IxIndexHandle> index_handle = ix_manager_->open_index(table_name, index_cols);
            ihs_.emplace(ix_manager_->get_index_name(table_name, index_cols), std::move(index_handle));
        }
    }
}

/**
 * @description: 把数据库相关的元数据刷入磁盘中
 */
void SmManager::flush_meta() {
    // 默认清空文件
    std::ofstream ofs(DB_META_NAME);
    ofs << db_;
}

/**
 * @description: 关闭数据库并把数据落盘
 */
// by march7 20240521
void SmManager::close_db() {
    // 1. 将数据库元数据刷入磁盘
    flush_meta();
    db_.name_.clear();
    db_.tabs_.clear();

    // 2. 关闭所有打开的记录文件和索引文件
    for (auto& file_entry : fhs_) {
        rm_manager_->close_file(file_entry.second.get());
    }
    fhs_.clear(); // 清空记录文件句柄映射

    for (auto& index_entry : ihs_) {
        ix_manager_->close_index(index_entry.second.get());
    }
    ihs_.clear(); // 清空索引文件句柄映射
}

/**
 * @description: 显示所有的表,通过测试需要将其结果写入到output.txt,详情看题目文档
 * @param {Context*} context 
 */
void SmManager::show_tables(Context* context) {
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);
    outfile << "| Tables |\n";
    RecordPrinter printer(1);
    printer.print_separator(context);
    printer.print_record({"Tables"}, context);
    printer.print_separator(context);
    for (auto &entry : db_.tabs_) {
        auto &tab = entry.second;
        printer.print_record({tab.name}, context);
        outfile << "| " << tab.name << " |\n";
    }
    printer.print_separator(context);
    outfile.close();
}

/**
 * @description: 显示表的元数据
 * @param {string&} tab_name 表名称
 * @param {Context*} context 
 */
void SmManager::desc_table(const std::string& tab_name, Context* context) {
    TabMeta &tab = db_.get_table(tab_name);

    std::vector<std::string> captions = {"Field", "Type", "Index"};
    RecordPrinter printer(captions.size());
    // Print header
    printer.print_separator(context);
    printer.print_record(captions, context);
    printer.print_separator(context);
    // Print fields
    for (auto &col : tab.cols) {
        std::vector<std::string> field_info = {col.name, coltype2str(col.type), col.index ? "YES" : "NO"};
        printer.print_record(field_info, context);
    }
    // Print footer
    printer.print_separator(context);
}

/**
 * @description: 创建表
 * @param {string&} tab_name 表的名称
 * @param {vector<ColDef>&} col_defs 表的字段
 * @param {Context*} context 
 */
void SmManager::create_table(const std::string& tab_name, const std::vector<ColDef>& col_defs, Context* context) {
    if (db_.is_table(tab_name)) {
        throw TableExistsError(tab_name);
    }
    // Create table meta
    int curr_offset = 0;
    TabMeta tab;
    tab.name = tab_name;
    for (auto &col_def : col_defs) {
        ColMeta col = {.tab_name = tab_name,
                       .name = col_def.name,
                       .type = col_def.type,
                       .len = col_def.len,
                       .offset = curr_offset,
                       .index = false};
        curr_offset += col_def.len;
        tab.cols.push_back(col);
    }
    // Create & open record file
    int record_size = curr_offset;  // record_size就是col meta所占的大小（表的元数据也是以记录的形式进行存储的）
    rm_manager_->create_file(tab_name, record_size);
    db_.tabs_[tab_name] = tab;
    // fhs_[tab_name] = rm_manager_->open_file(tab_name);
    fhs_.emplace(tab_name, rm_manager_->open_file(tab_name));

    flush_meta();
}

/**
 * @description: 删除表
 * @param {string&} tab_name 表的名称
 * @param {Context*} context
 */
// by zmjjkk 20240521
void SmManager::drop_table(const std::string& tab_name, Context* context) {
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }
    
    // delete index by march7 20240522
    TabMeta tabmeta = db_.get_table(tab_name);
    for (const auto& index : db_.get_table(tab_name).indexes) {
        // 删除索引文件
        std::string index_name = ix_manager_->get_index_name(tab_name, index.cols);
        ix_manager_->close_index(ihs_.at(index_name).get());
        ix_manager_->destroy_index(tab_name, index.cols);
        // 删除索引句柄
        ihs_.erase(index_name);
    }
    // 删除索引元数据
    tabmeta.indexes.clear();

    if(fhs_.find(tab_name)!=fhs_.end()) {
        // delete record file
        rm_manager_->close_file(fhs_[tab_name].get());
        fhs_.erase(tab_name);
    }
    rm_manager_->destroy_file(tab_name);
    // delete meta file
    db_.tabs_.erase(tab_name);

    flush_meta();
}

/**
 * @description: 创建索引
 * @param {string&} tab_name 表的名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
// by march7 20240522
void SmManager::create_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    // 1. 检查表是否存在
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    // 2. 检查索引是否已存在
    if (ix_manager_->exists(tab_name, col_names)) {
        throw IndexExistsError(tab_name, col_names);
    }

    // 3. 获取表的元数据
    TabMeta& tab_meta = db_.get_table(tab_name);

    // 4. 检查索引字段是否存在于表的列中
    for (const std::string& col_name : col_names) {
        bool found = false;
        for (const ColMeta& col_meta : tab_meta.cols) {
            if (col_meta.name == col_name) {
                found = true;
                break;
            }
        }
        if (!found) {
            throw ColumnNotFoundError(col_name);
        }
    }

    // 5. 准备索引列的元数据
    std::vector<ColMeta> index_cols;
    int col_tot_len = 0;
    int col_num = col_names.size();
    for (const std::string& col_name : col_names) {
        for (const ColMeta& col_meta : tab_meta.cols) {
            if (col_meta.name == col_name) {
                index_cols.push_back(col_meta);
                col_tot_len += col_meta.len; // 更新索引字段长度总和
                break;
            }
        }
    }

    // 6. 创建索引文件
    ix_manager_->create_index(tab_name, index_cols);

    // 7. 索引所有记录到索引文件中
    std::unique_ptr<IxIndexHandle> index_handle = ix_manager_->open_index(tab_name, index_cols);
    auto file_handle = fhs_[tab_name].get();

    // 扫描文件中的每条记录，并将其插入索引中
    for (RmScan rm_scan(file_handle); !rm_scan.is_end(); rm_scan.next()) {
        auto rec = file_handle->get_record(rm_scan.rid(), context); // 获取记录

        // 使用智能指针，并在每次循环中创建新的索引键
        std::vector<char> key(col_tot_len);

        int offset = 0;
        for (const auto& col_meta : index_cols) {
            // 提取索引字段值并组成索引键
            memcpy(key.data() + offset, rec->data + col_meta.offset, col_meta.len);
            offset += col_meta.len;
        }

        // 将索引键和记录的 RID 插入到索引中
        index_handle->insert_entry(key.data(), rm_scan.rid(), context->txn_);
    }
    // 8. 存储索引句柄
    ihs_.emplace(ix_manager_->get_index_name(tab_name, index_cols), std::move(index_handle));

    // 9. 添加索引元数据
    IndexMeta index_meta;
    index_meta.tab_name = tab_name;
    index_meta.col_tot_len = col_tot_len; // 设置索引字段长度总和
    index_meta.col_num = col_num; // 设置字段数量
    index_meta.cols = index_cols;
    tab_meta.indexes.push_back(index_meta);

    // 10. 刷新数据库元数据
    flush_meta();

    // todo trx
}

void SmManager::show_index(const std::string& tab_name, Context* context) {
    // 打开输出文件，以附加模式写入 by jinx 20240522
    std::fstream outfile;
    outfile.open("output.txt", std::ios::out | std::ios::app);

    // 检查表是否存在
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    // 获取表的元数据
    const TabMeta& tab_meta = db_.get_table(tab_name);

    // 初始化 RecordPrinter
    RecordPrinter printer(3);  // 三列：表名，是否唯一，列名

    // 遍历表的所有索引
    for (const auto& index_meta : tab_meta.indexes) {
        // 构建索引的列名列表
        std::string column_names = "(";
        for (size_t i = 0; i < index_meta.cols.size(); ++i) {
            column_names += index_meta.cols[i].name;
            if (i < index_meta.cols.size() - 1) {
                column_names += ",";
            }
        }
        column_names += ")";

        // 打印索引信息到输出文件和服务端
        std::vector<std::string> record = {tab_name, "unique", column_names};
        printer.print_record(record, context);
        outfile << "| " << tab_name << " | unique | " << column_names << " |\n";
    }

    // 关闭输出文件
    outfile.close();
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<string>&} col_names 索引包含的字段名称
 * @param {Context*} context
 */
// by march7 20240522
void SmManager::drop_index(const std::string& tab_name, const std::vector<std::string>& col_names, Context* context) {
    // 1. 检查表是否存在 buyong jiancha haha
     if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    // 2. 检查索引是否存在
    if (!ix_manager_->exists(tab_name, col_names)) {
        throw IndexNotFoundError(tab_name, col_names);
    }

    // 3. 获取表的元数据
    TabMeta& tab_meta = db_.get_table(tab_name);

    // 4. 删除索引文件
    std::string index_name = ix_manager_->get_index_name(tab_name, col_names);
    ix_manager_->close_index(ihs_.at(index_name).get());
    ix_manager_->destroy_index(tab_name, col_names);

    // 5. 删除索引句柄
    ihs_.erase(index_name);

    // 6. 删除索引元数据
    auto it = tab_meta.get_index_meta(col_names);
    tab_meta.indexes.erase(it);

    // 7. 刷新数据库元数据
    flush_meta();
}

/**
 * @description: 删除索引
 * @param {string&} tab_name 表名称
 * @param {vector<ColMeta>&} 索引包含的字段元数据
 * @param {Context*} context
 */
// by march7 20240522
void SmManager::drop_index(const std::string& tab_name, const std::vector<ColMeta>& cols, Context* context) {
    // 1. 检查表是否存在
    if (!db_.is_table(tab_name)) {
        throw TableNotFoundError(tab_name);
    }

    // 2. 检查索引是否存在
    std::vector<std::string> col_names;
    for (const auto& col : cols) {
        col_names.push_back(col.name);
    }

    drop_index(tab_name, col_names, context);
}